//
// SKTArc.m
// By mkino
//

#import "SKTArc.h"

@implementation SKTArc

//
// Initialization
//

- (id)init {
    self = [super init];
    if (self) {
        _direction = NoneDirection;
    }
    return self;
}

- (id)copyWithZone:(NSZone *)zone {
    id newObj = [super copyWithZone:zone];
    
    [newObj setDirection:[self direction]];
    
    return newObj;
}

//
// Accessors
//

- (void)setDirection:(int)direction {
    _direction = direction;
}

- (int)direction {
    return _direction;
}

//
// Extended mutation
//

- (void)flipHorizontally {
    _direction *= -1;
    
    return;
}

- (void)flipVertically {
    if (abs(_direction) == 1) {
        _direction *= 2;
    }
    else {
        _direction /= 2;
    }
    
    return;
}

- (void)makeNaturalSize {
    NSRect bounds = [self bounds];
    if (bounds.size.width < bounds.size.height) {
        bounds.size.height = bounds.size.width;
        [self setBounds:bounds];
    } else if (bounds.size.width > bounds.size.height) {
        bounds.size.width = bounds.size.height;
        [self setBounds:bounds];
    }
}

//
// Persistence
//

NSString *SKTArcDirectionKey = @"ArcDirect";

- (NSMutableDictionary *)propertyListRepresentation {
    NSMutableDictionary *dict = [super propertyListRepresentation];
    [dict setObject:[NSNumber numberWithInt:[self direction]] forKey:SKTArcDirectionKey];
    return dict;
}

- (void)loadPropertyListRepresentation:(NSDictionary *)dict {
    id obj;
    
    [super loadPropertyListRepresentation:dict];

    obj = [dict objectForKey:SKTArcDirectionKey];
    if (obj) {
        [self setDirection:[obj intValue]];
    }
}

//
// Drawing
//

- (void)bezierPathForStroke:(NSBezierPath**)pathForStroke andPathForFill:(NSBezierPath**)pathForFill {
    NSRect	bounds = [self bounds];
    NSPoint	fromPt, toPt, ctrl1Pt, ctrl2Pt, centerPt;
    
    fromPt = toPt = ctrl1Pt = ctrl2Pt = centerPt = NSZeroPoint;
    *pathForStroke = nil;
    *pathForFill = nil;
    
    // In case of creation
    if ([self isCreating]) {
        _direction = [self directionOfCreating];
    }
    
    // Create points for bezier paths
    switch (_direction) {
    case UpperRightDirection: {
        // Upper right
        fromPt.x = bounds.origin.x;
        fromPt.y = bounds.origin.y;
        toPt.x = bounds.origin.x + bounds.size.width;
        toPt.y = bounds.origin.y + bounds.size.height;
        ctrl1Pt.x = fromPt.x + bounds.size.width / 2.0;
        ctrl1Pt.y = fromPt.y;
        ctrl2Pt.x = toPt.x;
        ctrl2Pt.y = toPt.y - bounds.size.height / 2.0;
        
        break;
    }
    case LowerRightDirection: {
        // Lower right
        fromPt.x = bounds.origin.x;
        fromPt.y = bounds.origin.y + bounds.size.height;
        toPt.x = bounds.origin.x + bounds.size.width;
        toPt.y = bounds.origin.y;
        ctrl1Pt.x = fromPt.x + bounds.size.width / 2.0;
        ctrl1Pt.y = fromPt.y;
        ctrl2Pt.x = toPt.x;
        ctrl2Pt.y = toPt.y + bounds.size.height / 2.0;
        
        break;
    }
    case UpperLeftDirection: {
        fromPt.x = bounds.origin.x + bounds.size.width;
        fromPt.y = bounds.origin.y;
        toPt.x = bounds.origin.x;
        toPt.y = bounds.origin.y + bounds.size.height;
        ctrl1Pt.x = fromPt.x - bounds.size.width / 2.0;
        ctrl1Pt.y = fromPt.y;
        ctrl2Pt.x = toPt.x;
        ctrl2Pt.y = toPt.y - bounds.size.height / 2.0;
        
        break;
    }
    case LowerLeftDirection: {
        fromPt.x = bounds.origin.x + bounds.size.width;
        fromPt.y = bounds.origin.y + bounds.size.height;
        toPt.x = bounds.origin.x;
        toPt.y = bounds.origin.y;
        ctrl1Pt.x = fromPt.x - bounds.size.width / 2.0;
        ctrl1Pt.y = fromPt.y;
        ctrl2Pt.x = toPt.x;
        ctrl2Pt.y = toPt.y + bounds.size.height / 2.0;
        
        break;
    }
    default : {
        // Return nil paths
        return;
    }
    }
    
    centerPt.x = fromPt.x;
    centerPt.y = toPt.y;
    
    // Create the path for stroking
    *pathForStroke = [NSBezierPath bezierPath];
    [*pathForStroke moveToPoint:fromPt];
    [*pathForStroke curveToPoint:toPt controlPoint1:ctrl1Pt controlPoint2:ctrl2Pt];
    [*pathForStroke setLineWidth:[self strokeLineWidth]];
    
    // Create the path for filling
    *pathForFill = [NSBezierPath bezierPath];
    [*pathForFill moveToPoint:centerPt];
    [*pathForFill lineToPoint:fromPt];
    [*pathForFill curveToPoint:toPt controlPoint1:ctrl1Pt controlPoint2:ctrl2Pt];
    [*pathForFill closePath];
}

- (void)drawInView:(SKTGraphicView *)view isSelected:(BOOL)flag {
    NSBezierPath*	pathForStroke;
    NSBezierPath*	pathForFill;
	[self bezierPathForStroke:&pathForStroke andPathForFill:&pathForFill];
    
    if (pathForStroke && pathForFill) {
        if ([self drawsFill]) {
            [[self fillColor] set];
            [pathForFill fill];
        }
        if ([self drawsStroke]) {
            [[self strokeColor] set];
            [pathForStroke stroke];
        }
    }
    if (flag) {
        [self drawHandlesInView:view];
    }
}

//
// Event handling
//

- (BOOL)hitTest:(NSPoint)point isSelected:(BOOL)isSelected {
    if (isSelected && ([self knobUnderPoint:point] != NoKnob)) {
        return YES;
    } else {
        NSBezierPath*	pathForStroke;
        NSBezierPath*	pathForFill;
        [self bezierPathForStroke:&pathForStroke andPathForFill:&pathForFill];

        if (pathForFill) {
            if ([pathForFill containsPoint:point]) {
                return YES;
            }
        } else {
            if (NSPointInRect(point, [self bounds])) {
                return YES;
            }
        }
        return NO;
    }
}

@end
