//
// SKTBezierPath.m
// By mkino
//

#import "SKTDrawDocument.h"
#import "SKTGraphicView.h"
#import "SKTBezierPath.h"

static float	SKTBezierPathHandleWidthHalf = 2.0;
static NSPoint	SktInvalidPoint = { -9999.0f, -9999.0f };

@implementation SKTBezierPath

//
// Creation
//

- (id)init {
    self = [super init];
    if (self) {
        _points = [[NSMutableArray array] retain];
        _isPathClosed = NO;
        _transform = [[NSAffineTransform transform] retain];
        
        _isFlippedVertically = NO;
        _isFlippedHorizontally = NO;
        _selectedPointIndex = -1;
    }
    
    return self;
}

- (void)dealloc {
    [_points release];
    [_transform release];
	[super dealloc];
}

- (id)copyWithZone:(NSZone*)zone {
    id	newObj = [super copyWithZone:zone];
    
    [newObj setPoints:_points];
    [newObj setPathClosed:_isPathClosed];
    [newObj setTransform:_transform];
    
    return newObj;
}

//
// Accessors
//

- (void)setPoints:(NSArray*)points {
    [_points removeAllObjects];
    [_points addObjectsFromArray:points];
}

- (NSArray*)points {
    return _points;
}

- (void)setPathClosed:(BOOL)isPathClosed {
    _isPathClosed = isPathClosed;
}

- (BOOL)isPathClosed {
    return _isPathClosed;
}

- (void)setTransform:(NSAffineTransform*)transform {
    [_transform release];
    _transform = [transform retain];
}

- (NSAffineTransform*)transform {
    return _transform;
}

//
// Anchor points and control points management
//

- (int)numberOfAnchorPoint {
    int	count = [_points count];
    if (count == 0) {
        return 0;
    }
    
    return ([_points count] - 1) / 3 + 1;
}

- (int)indexOfPoint:(NSPoint)point {
    int	i;
    for (i = 0; i < [_points count]; i++) {
        if (NSEqualPoints(point, [[_points objectAtIndex:i] pointValue])) {
            return i;
        }
    }
    
    return -1;
}

- (int)indexOfLastAnchorPoint {
    return [_points count] - 1;
}

- (NSPoint)pointAtIndex:(int)index {
    return [[_points objectAtIndex:index] pointValue];
}

- (void)appendFirstAnchorPoint:(NSPoint)point
{
    [_points addObject:[NSValue valueWithPoint:point]];
}

- (void)appendAnchorPoint:(NSPoint)point controlPoint1:(NSPoint)control1 controlPoint2:(NSPoint)control2
{
    [_points addObject:[NSValue valueWithPoint:control1]];
    [_points addObject:[NSValue valueWithPoint:point]];
    [_points addObject:[NSValue valueWithPoint:control2]];
}

- (void)movePointAtIndex:(int)index toPoint:(NSPoint)point {
    [_points replaceObjectAtIndex:index withObject:[NSValue valueWithPoint:point]];
}

- (void)movePointAtIndex:(int)index bySize:(NSSize)size {
    NSPoint	point = [self pointAtIndex:index];
    point.x += size.width;
    point.y += size.height;
    [self movePointAtIndex:index toPoint:point];
}

- (void)removeAnchorPointAtIndex:(int)index {
    int	lastIndex;
    
    if (![self isAnchorPoint:index]) {
        return;
    }
    
    lastIndex = [self indexOfLastAnchorPoint];
    if (index == 0 || index == lastIndex) {
        if(_isPathClosed) {
            [_points replaceObjectAtIndex:lastIndex withObject:[_points objectAtIndex:3]];
            [_points replaceObjectAtIndex:lastIndex - 1 withObject:[_points objectAtIndex:2]];
            [_points removeObjectsInRange:NSMakeRange(0, 3)];
        }
        else {
            if (index == 0) {
                [_points removeObjectsInRange:NSMakeRange(0, 3)];
            }
            else if (index == lastIndex) {
                [_points removeObjectsInRange:NSMakeRange(lastIndex - 2, 3)];
            }
        }
    }
    else {
        [_points removeObjectsInRange:NSMakeRange(index - 1, 3)];
    }
}

- (BOOL)isAnchorPoint:(int)index {
    return index % 3 == 0;
}

//
// Extended mutation
//

- (void)flipHorizontally {
    _isFlippedVertically = ~_isFlippedVertically;
    
    return;
}

- (void)flipVertically {
    _isFlippedHorizontally = ~_isFlippedHorizontally;
    
    return;
}

//
// Persistence
//

NSString*	SKTBezierPathPointsKey = @"BezierPoints";
NSString*	SKTBezierPathIsPathClosedKey = @"BezierIsPathClosed";

- (NSMutableDictionary *)propertyListRepresentation {
    NSMutableDictionary*	dict = [super propertyListRepresentation];
    NSMutableArray*			array = [NSMutableArray arrayWithCapacity:[_points count]];
    int						i;
    
    // Apply transform at first
    [self applyTransform];
    
    // Make an anchor points array represented in NSString
    for (i = 0; i < [_points count]; i++) {
        [array addObject:NSStringFromPoint([self pointAtIndex:i])];
    }
    [dict setObject:array forKey:SKTBezierPathPointsKey];
    
    // For _isPathClosed
    [dict setObject:[NSNumber numberWithBool:_isPathClosed] forKey:SKTBezierPathIsPathClosedKey];
    
    return dict;
}

- (void)loadPropertyListRepresentation:(NSDictionary *)dict {
    id	obj;
    int	i;
    
    [super loadPropertyListRepresentation:dict];
    
    obj = [dict objectForKey:SKTBezierPathPointsKey];
    if (obj) {
        NSMutableArray*	array = [NSMutableArray arrayWithCapacity:[obj count]];
        for (i = 0; i < [obj count]; i++) {
            [array addObject:[NSValue valueWithPoint:NSPointFromString([obj objectAtIndex:i])]];
        }
        [self setPoints:array];
    }
    obj = [dict objectForKey:SKTBezierPathIsPathClosedKey];
    if (obj) {
        [self setPathClosed:[obj intValue]];
    }
}

//
// Drawing
//

- (NSRect)drawingBounds {
    NSRect	controlBounds, rect;
    float	inset = -SKTBezierPathHandleWidthHalf;
    if ([self drawsStroke]) {
        float halfLineWidth = ([self strokeLineWidth] / 2.0) + 1.0;
        if (-halfLineWidth < inset) {
            inset = -halfLineWidth;
        }
    }
    inset += -1.0;
    
    if (![self isCreating] && ![self isEditing]) {
        return NSInsetRect([[self bezierPath] bounds], inset, inset);
    }
    
    controlBounds = [[self bezierPath] controlPointBounds];
    if (controlBounds.size.width == 0) {
        controlBounds.size.width = 1;
    }
    if (controlBounds.size.height == 0) {
        controlBounds.size.height = 1;
    }
    rect.origin.x = _lastControlPoint.x - SKTBezierPathHandleWidthHalf;
    rect.origin.y = _lastControlPoint.y - SKTBezierPathHandleWidthHalf;
    rect.size.width = SKTBezierPathHandleWidthHalf * 2 + 1.0;
    rect.size.height = SKTBezierPathHandleWidthHalf * 2 + 1.0;
    
    return NSInsetRect(NSUnionRect(controlBounds, rect), inset, inset);
}

- (NSBezierPath*)bezierPath {
    NSBezierPath*	path = [self origBezierPath];
    
    NSRect			bounds = [self bounds];
    NSRect			bezierBounds = [path bounds];
    float			scaleX = 0.0f, scaleY = 0.0f;
    
    if (![self isCreating] && 
        ![self isEditing] && 
        ([self isMoving] || [self isSizeChanging])) 
    {
        [_transform setTransformStruct:SKTNormalizeStruct];
        
        // For moving
        [_transform translateXBy:bounds.origin.x yBy:bounds.origin.y];
        
        // For size changing
        if (bezierBounds.size.width != 0) {
            scaleX = bounds.size.width / bezierBounds.size.width;
        }
        if (bezierBounds.size.height != 0) {
            scaleY = bounds.size.height / bezierBounds.size.height;
        }
        [_transform scaleXBy:scaleX yBy:scaleY];
        
        // Back to origin
        [_transform translateXBy:-bezierBounds.origin.x yBy:-bezierBounds.origin.y];
        
        // For vertical flip
        if (_isFlippedVertically) {
            [_transform prependTransform:SKTVerticalFlipTransform];
            [_transform translateXBy:-(2 * bezierBounds.origin.x + bezierBounds.size.width) yBy:0.0];            
        }
        // For horizontal flip
        if (_isFlippedHorizontally) {
            [_transform prependTransform:SKTHorizontalFlipTransform];
            [_transform translateXBy:0.0 yBy:-(2 * bezierBounds.origin.y + bezierBounds.size.height)];            
        }
    }
    
    return [_transform transformBezierPath:path];
}

- (NSBezierPath*)origBezierPath {
    NSBezierPath*	path = [NSBezierPath bezierPath];
    NSPoint			anchor1, control1, control2, anchor2;
    int	numberOfAnchorPoint = [self numberOfAnchorPoint];
    int	i, index = 0;
    
    if (numberOfAnchorPoint >= 1) {
        anchor1 = [self pointAtIndex:index++];
        [path moveToPoint:anchor1];
        
        if (numberOfAnchorPoint >= 2) {
            for (i = 1; i < numberOfAnchorPoint; i++) {
                control1 = [self pointAtIndex:index++];
                control2 = [self pointAtIndex:index++];
                anchor2 = [self pointAtIndex:index++];
                
                if (NSEqualPoints(control1, SktInvalidPoint)) {
                    control1 = anchor1;
                }
                if (NSEqualPoints(control2, SktInvalidPoint)) {
                    control2 = anchor2;
                }
                
                [path curveToPoint:anchor2 controlPoint1:control1 controlPoint2:control2];
                anchor1 = anchor2;
            }
        }
        
        if (_isPathClosed) {
            [path closePath];
        }
        
    }
    
    [path setLineWidth:[self strokeLineWidth]];
    
    return path;
}

- (void)drawInView:(SKTGraphicView *)view isSelected:(BOOL)flag {
    NSBezierPath *path = [self bezierPath];
    if (path) {
        if ([self drawsFill]) {
            [[self fillColor] set];
            [path fill];
        }
        if ([self drawsStroke]) {
            [[self strokeColor] set];
            [path stroke];
        }
    }
    if (flag) {
        [self drawHandlesInView:view];
    }
    
    // Handles for creating or editing
    if ([self isCreating] || [self isEditing]) {
        [self drawAnchorHandlesInView:view];
        [self drawControlHandlesInView:view];
    }
}

- (void)drawAnchorHandlesInView:(SKTGraphicView*)view {
    int	i;
    NSPoint	point;
    NSRect	rect;
    BOOL	isCreating = [self isCreating];
    BOOL	isEditing = [self isEditing];
    
    for (i = 0; i < [_points count]; i++) {
        if (i != 0 && i % 3 != 0) {
            continue;
        }
        
        point = [self pointAtIndex:i];
        
        if (isCreating) {
            rect.origin.x = point.x - SKTBezierPathHandleWidthHalf;
            rect.origin.y = point.y - SKTBezierPathHandleWidthHalf;
            rect.size.width = SKTBezierPathHandleWidthHalf * 2 + 1.0;
            rect.size.height = SKTBezierPathHandleWidthHalf * 2 + 1.0;
            
            [[NSColor blackColor] set];
            NSRectFill(rect);
        }
        else if (isEditing) {
            rect.origin.x = point.x - SKTBezierPathHandleWidthHalf;
            rect.origin.y = point.y - SKTBezierPathHandleWidthHalf;
            rect.size.width = SKTBezierPathHandleWidthHalf * 2 + 1.0;
            rect.size.height = SKTBezierPathHandleWidthHalf * 2 + 1.0;
            
            [[NSColor whiteColor] set];
            NSRectFill(rect);
            
            [[NSColor blackColor] set];
            NSFrameRect(rect);
            
            if (i == _selectedPointIndex) {
                NSFrameRect(NSInsetRect(rect, 1.0, 1.0));
            }
        }
    }
}

- (void)drawControlHandlesInView:(SKTGraphicView*)view {
    NSPoint	anchorPoint, ctrlPoint, ctrlPoint1;
    int		lastAnchorIndex = [self indexOfLastAnchorPoint];
    
    if ([self isCreating]) {
        anchorPoint = [self pointAtIndex:lastAnchorIndex];
        
        [self drawControlHandlesFromAnchorPoint:anchorPoint toControlPoint:_lastControlPoint inView:view];
        
        if (lastAnchorIndex != 0) {
            ctrlPoint = [self pointAtIndex:lastAnchorIndex - 1];
            [self drawControlHandlesFromAnchorPoint:anchorPoint toControlPoint:ctrlPoint inView:view];
        }
    }
    else if ([self isEditing]) {
        if (_selectedPointIndex < 0 || [_points count] == 1) {
            return;
        }
        
        anchorPoint	= [self pointAtIndex:_selectedPointIndex];
        if (_selectedPointIndex == 0) {
            ctrlPoint = [self pointAtIndex:1];
            if (_isPathClosed) {
                ctrlPoint1 = [self pointAtIndex:lastAnchorIndex - 1];
            }
            else {
                ctrlPoint1 = SktInvalidPoint;
            }
        }
        else if (_selectedPointIndex == lastAnchorIndex) {
            if (_isPathClosed) {
                ctrlPoint = [self pointAtIndex:1];
            }
            else {
                ctrlPoint = SktInvalidPoint;
            }
            ctrlPoint1 = [self pointAtIndex:lastAnchorIndex - 1];
        }
        else {
            ctrlPoint = [self pointAtIndex:_selectedPointIndex - 1];
            ctrlPoint1 = [self pointAtIndex:_selectedPointIndex + 1];
        }
        
        if (!NSEqualPoints(ctrlPoint, SktInvalidPoint)) {
            [self drawControlHandlesFromAnchorPoint:anchorPoint toControlPoint:ctrlPoint inView:view];
        }
        if (!NSEqualPoints(ctrlPoint1, SktInvalidPoint)) {
            [self drawControlHandlesFromAnchorPoint:anchorPoint toControlPoint:ctrlPoint1 inView:view];
        }
    }
}

- (void)drawControlHandlesFromAnchorPoint:(NSPoint)anchorPoint toControlPoint:(NSPoint)controlPoint inView:(SKTGraphicView*)view {
    NSRect	rect;
    
    [[NSColor blueColor] set];
    
    [NSBezierPath strokeLineFromPoint:anchorPoint toPoint:controlPoint];
    
    rect.origin.x = controlPoint.x - SKTBezierPathHandleWidthHalf;
    rect.origin.y = controlPoint.y - SKTBezierPathHandleWidthHalf;
    rect.size.width = SKTBezierPathHandleWidthHalf * 2 + 1.0;
    rect.size.height = SKTBezierPathHandleWidthHalf * 2 + 1.0;
    [[NSBezierPath bezierPathWithOvalInRect:rect] fill];
}

//
// Event handling
//

- (BOOL)createWithEvent:(NSEvent *)event inView:(SKTGraphicView *)view {
	[self setLineArrowMode:[view lineArrowMode]];
	[self setFillColor:[view fillColor]];
	[self setDrawsFill:[view drawsFill]];
	[self setStrokeColor:[view strokeColor]];
	[self setDrawsStroke:[view drawsStroke]];
	[self setStrokeLineWidth:[view strokeLineWidth]];
    NSPoint	point = [view convertPoint:[event locationInWindow] fromView:nil];
    NSPoint	lastAnchorPt;
    NSRect	drawingBounds, lastDrawingBounds;
    BOOL	snapsToGrid = [view snapsToGrid];
    float	spacing = [view gridSpacing];
    BOOL	echoToRulers = [[view enclosingScrollView] rulersVisible];
    NSEventType	type;
    
    [self startBoundsManipulation];
    if (snapsToGrid) {
        point.x = floor((point.x / spacing) + 0.5) * spacing;
        point.y = floor((point.y / spacing) + 0.5) * spacing;
    }
    [self setBounds:NSMakeRect(point.x, point.y, 0.0, 0.0)];
    if (echoToRulers) {
        [view beginEchoingMoveToRulers:[self bounds]];
    }
    
    // Append first anchor point
    [self appendFirstAnchorPoint:point];
    lastAnchorPt = point;
    
    lastDrawingBounds.origin = point;
    lastDrawingBounds.size.width = 1;
    lastDrawingBounds.size.height = 1;
    
    // Tracking loop
    while (1) {
        event = [[view window] nextEventMatchingMask:(NSLeftMouseDownMask | NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        point = [view convertPoint:[event locationInWindow] fromView:nil];
        if (snapsToGrid) {
            point.x = floor((point.x / spacing) + 0.5) * spacing;
            point.y = floor((point.y / spacing) + 0.5) * spacing;
        }
        
        type = [event type];
        if (type == NSLeftMouseDown && [event clickCount] == 1) {
            // Append points for a bezier path
            [self appendAnchorPoint:point controlPoint1:_lastControlPoint controlPoint2:point];
            
            lastAnchorPt = point;
            _lastControlPoint = point;
        }
        else if (type == NSLeftMouseDragged) {
            int		index = [self indexOfLastAnchorPoint];
            NSPoint	flippedCtrlPt;
            flippedCtrlPt.x = 2 * lastAnchorPt.x - point.x;
            flippedCtrlPt.y = 2 * lastAnchorPt.y - point.y;
            
            // Move the last control point
            if (index != 0) {
                [self movePointAtIndex:index - 1 toPoint:flippedCtrlPt];
            }
            _lastControlPoint = point;
        }
        else if (type == NSLeftMouseUp) {
            if ([_points count] > 1) {
                int	index = [self pointUnderPoint:lastAnchorPt];
                if (index == 0) {
                    // Fit to first point
                    [self movePointAtIndex:[self indexOfLastAnchorPoint] toPoint:[self pointAtIndex:0]];
                    
                    // Close path and go out from loop
                    _isPathClosed = YES;
                    break;
                }
            }
        }
        
        // Set bounds and calculate drawing bounds
        [self setBounds:[[self bezierPath] bounds]];
        drawingBounds = [self drawingBounds];
        [view setNeedsDisplayInRect:NSUnionRect(drawingBounds, lastDrawingBounds)];
        lastDrawingBounds = drawingBounds;
        
        if (echoToRulers) {
            [view continueEchoingMoveToRulers:[self bounds]];
        }
        
        // In case of double click
        if ([event clickCount] >= 2) {
            break;
        }
    }
    if (echoToRulers) {
        [view stopEchoingMoveToRulers];
    }
    
    [self stopBoundsManipulation];
    
    return !NSEqualRects([self bounds], NSZeroRect);
}

- (BOOL)isEditable {
    return YES;
}

- (void)startEditingWithEvent:(NSEvent *)event inView:(SKTGraphicView *)view {
    [super startEditingWithEvent:event inView:view];
    
    [self applyTransform];
    [view setEditingGraphic:self editorView:nil];
    [view setNeedsDisplayInRect:[self drawingBounds]];
    
    return;
}

- (void)endEditingInView:(SKTGraphicView *)view {
    [view setNeedsDisplayInRect:[self drawingBounds]];
    [super endEditingInView:view];
    
    _selectedPointIndex = -1;
    
    return;
}

- (BOOL)editWithEvent:(NSEvent*)event inView:(SKTGraphicView*)view {
    NSPoint	point = [view convertPoint:[event locationInWindow] fromView:nil];
    NSPoint	oldPoint = point;
    NSSize	diffSize;
    NSRect	drawingBounds, lastDrawingBounds;
    int		pointIndex = [self pointUnderPoint:point];
    BOOL	snapsToGrid = [view snapsToGrid];
    float	spacing = [view gridSpacing];
    BOOL	echoToRulers = [[view enclosingScrollView] rulersVisible];
    
    if (pointIndex < 0) {
        return NO;
    }
    
    if ([self isAnchorPoint:pointIndex]) {
        _selectedPointIndex = pointIndex;
    }
    else {
        if (_selectedPointIndex < 0) {
            return NO;
        }
    }
    
    lastDrawingBounds = [self drawingBounds];
    
    while (1) {
        event = [[view window] nextEventMatchingMask:(NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        
        if ([event type] == NSLeftMouseDragged) {
            point = [view convertPoint:[event locationInWindow] fromView:nil];
            if (snapsToGrid) {
                point.x = floor((point.x / spacing) + 0.5) * spacing;
                point.y = floor((point.y / spacing) + 0.5) * spacing;
            }
            
            [self movePointAtIndex:pointIndex toPoint:point];
            if ([self isAnchorPoint:pointIndex]) {
                int	index = [self indexOfLastAnchorPoint];
                diffSize.width = point.x - oldPoint.x;
                diffSize.height = point.y - oldPoint.y;
                
                if (_isPathClosed && (pointIndex == 0 || pointIndex == index)) {
                    [self movePointAtIndex:0 toPoint:point];
                    [self movePointAtIndex:index toPoint:point];
                    [self movePointAtIndex:1 bySize:diffSize];
                    [self movePointAtIndex:index - 1 bySize:diffSize];
                }
                else {
                    if (pointIndex > 0) {
                        [self movePointAtIndex:pointIndex - 1 bySize:diffSize];
                    }
                    if (pointIndex < index) {
                        [self movePointAtIndex:pointIndex + 1 bySize:diffSize];
                    }
                }
            }
        }
        oldPoint = point;
        
        [self setBounds:[[self bezierPath] bounds]];
        drawingBounds = [self drawingBounds];
        [view setNeedsDisplayInRect:NSUnionRect(drawingBounds, lastDrawingBounds)];
        lastDrawingBounds = drawingBounds;
        
        if (echoToRulers) {
            [view continueEchoingMoveToRulers:[self bounds]];
        }
        
        if ([event type] == NSLeftMouseUp) {
            break;
        }
    }
    
    return YES;
}

- (void)deleteForEditingInView:(SKTGraphicView*)view {
    NSRect	drawingBounds;
    
    if (_selectedPointIndex != -1) {
        // If the points is last one
        if ([_points count] == 1) {
            [[self document] removeGraphic:self];
            return;
        }
        
        drawingBounds = [self drawingBounds];
        
        [self removeAnchorPointAtIndex:_selectedPointIndex];
        _selectedPointIndex = -1;
        
        [self setBounds:[[self origBezierPath] bounds]];
        drawingBounds = NSUnionRect(drawingBounds, [self drawingBounds]);
        [view setNeedsDisplayInRect:drawingBounds];
    }
}

//
// Bezier path specific
//

- (int)pointUnderPoint:(NSPoint)point {
    int	i;
    NSPoint	pt;
    NSRect	rect;
    
    for (i = 0; i < [_points count]; i++) {
        pt = [self pointAtIndex:i];
        pt = [_transform transformPoint:pt];
        
        rect.origin.x = pt.x - (SKTBezierPathHandleWidthHalf + 1);
        rect.origin.y = pt.y - (SKTBezierPathHandleWidthHalf + 1);
        rect.size.width = (SKTBezierPathHandleWidthHalf + 1) * 2 + 1.0;
        rect.size.height = (SKTBezierPathHandleWidthHalf + 1) * 2 + 1.0;
        
        if (NSPointInRect(point, rect)) {
            return i;
        }
    }
    
    return -1;
}

- (void)applyTransform {
    NSMutableArray*	points = [NSMutableArray arrayWithCapacity:[_points count]];
    NSEnumerator*	enumerator = [_points objectEnumerator];
    id				object;
    NSPoint			point;
    
    while (object = [enumerator nextObject]) {
        point = [object pointValue];
        point = [_transform transformPoint:point];
        [points addObject:[NSValue valueWithPoint:point]];
    }
    
    // Swap anchor points
    [_points release];
    _points = [points retain];
    
    // Normarize transform
    [_transform setTransformStruct:SKTNormalizeStruct];
    
    _isFlippedVertically = NO;
    _isFlippedHorizontally = NO;
}

@end
