//
// SKTPolygon.m
// By mkino
//

#import "SKTDrawDocument.h"
#import "SKTGraphicView.h"
#import "SKTPolygon.h"

static float	SKTPolygonHandleWidthHalf = 2.0;

@implementation SKTPolygon

//
// Creation
//

- (id)init {
    self = [super init];
    if (self) {
        _anchorPoints = [[NSMutableArray array] retain];
        _isPathClosed = NO;
        _transform = [[NSAffineTransform transform] retain];
        
        _isFlippedVertically = NO;
        _isFlippedHorizontally = NO;
        _selectedPointIndex = -1;
    }
    
    return self;
}

- (void)dealloc {
    [_anchorPoints release];
    [_transform release];
	[super dealloc];
}

- (id)copyWithZone:(NSZone*)zone {
    id	newObj = [super copyWithZone:zone];
    
    [newObj setAnchorPoints:_anchorPoints];
    [newObj setPathClosed:_isPathClosed];
    [newObj setTransform:_transform];
    
    return newObj;
}

//
// Accessors
//

- (void)setAnchorPoints:(NSArray*)anchorPoints {
    [_anchorPoints removeAllObjects];
    [_anchorPoints addObjectsFromArray:anchorPoints];
}

- (NSArray*)anchorPoints {
    return _anchorPoints;
}

- (void)setPathClosed:(BOOL)isPathClosed {
    _isPathClosed = isPathClosed;
}

- (BOOL)isPathClosed {
    return _isPathClosed;
}

- (void)setTransform:(NSAffineTransform*)transform {
    [_transform release];
    _transform = [transform retain];
}

- (NSAffineTransform*)transform {
    return _transform;
}

//
// Anchor points management
//

- (int)indexOfPoint:(NSPoint)point
{
    int	i;
    for (i = 0; i < [_anchorPoints count]; i++) {
        if (NSEqualPoints(point, [[_anchorPoints objectAtIndex:i] pointValue])) {
            return i;
        }
    }
    
    return -1;
}

- (NSPoint)pointAtIndex:(int)index {
    return [[_anchorPoints objectAtIndex:index] pointValue];
}

- (void)appendPoint:(NSPoint)point {
    [_anchorPoints addObject:[NSValue valueWithPoint:point]];
}

- (void)movePointAtIndex:(int)index toPoint:(NSPoint)point {
    [_anchorPoints replaceObjectAtIndex:index withObject:[NSValue valueWithPoint:point]];
}

- (void)removePointAtIndex:(int)index {
    [_anchorPoints removeObjectAtIndex:index];
}

//
// Extended mutation
//

- (void)flipHorizontally {
    _isFlippedVertically = ~_isFlippedVertically;
    
    return;
}

- (void)flipVertically {
    _isFlippedHorizontally = ~_isFlippedHorizontally;
    
    return;
}

//
// Persistence
//

NSString*	SKTPolygonAnchorPointsKey = @"PolyAnchorPoints";
NSString*	SKTPolygonIsPathClosedKey = @"PolyIsPathClosed";

- (NSMutableDictionary *)propertyListRepresentation {
    NSMutableDictionary*	dict = [super propertyListRepresentation];
    NSMutableArray*			array = [NSMutableArray arrayWithCapacity:[_anchorPoints count]];
    int						i;
    
    // Apply transform at first
    [self applyTransform];
    
    // Make an anchor points array represented in NSString
    for (i = 0; i < [_anchorPoints count]; i++) {
        [array addObject:NSStringFromPoint([self pointAtIndex:i])];
    }
    [dict setObject:array forKey:SKTPolygonAnchorPointsKey];
    
    // For _isPathClosed
    [dict setObject:[NSNumber numberWithBool:_isPathClosed] forKey:SKTPolygonIsPathClosedKey];
    
    return dict;
}

- (void)loadPropertyListRepresentation:(NSDictionary *)dict {
    id	obj;
    int	i;
    
    [super loadPropertyListRepresentation:dict];
    
    obj = [dict objectForKey:SKTPolygonAnchorPointsKey];
    if (obj) {
        NSMutableArray*	array = [NSMutableArray arrayWithCapacity:[obj count]];
        for (i = 0; i < [obj count]; i++) {
            [array addObject:[NSValue valueWithPoint:NSPointFromString([obj objectAtIndex:i])]];
        }
        [self setAnchorPoints:array];
    }
    obj = [dict objectForKey:SKTPolygonIsPathClosedKey];
    if (obj) {
        [self setPathClosed:[obj intValue]];
    }
}

//
// Drawing
//

- (NSBezierPath*)bezierPath {
    NSBezierPath*	path = [self origBezierPath];
    NSRect			bounds = [self bounds];
    NSRect			bezierBounds = [path bounds];
    float			scaleX = 0.0f, scaleY = 0.0f;
    
    if (![self isCreating] && ([self isMoving] || [self isSizeChanging])) {
        [_transform setTransformStruct:SKTNormalizeStruct];
        
        // For moving
        [_transform translateXBy:bounds.origin.x yBy:bounds.origin.y];
        
        // For size changing
        if (bezierBounds.size.width != 0) {
            scaleX = bounds.size.width / bezierBounds.size.width;
        }
        if (bezierBounds.size.height != 0) {
            scaleY = bounds.size.height / bezierBounds.size.height;
        }
        [_transform scaleXBy:scaleX yBy:scaleY];
        
        // Back to origin
        [_transform translateXBy:-bezierBounds.origin.x yBy:-bezierBounds.origin.y];
        
        // For vertical flip
        if (_isFlippedVertically) {
            [_transform prependTransform:SKTVerticalFlipTransform];
            [_transform translateXBy:-(2 * bezierBounds.origin.x + bezierBounds.size.width) yBy:0.0];            
        }
        // For horizontal flip
        if (_isFlippedHorizontally) {
            [_transform prependTransform:SKTHorizontalFlipTransform];
            [_transform translateXBy:0.0 yBy:-(2 * bezierBounds.origin.y + bezierBounds.size.height)];            
        }
    }
    
    return [_transform transformBezierPath:path];
}

- (NSBezierPath*)origBezierPath {
    NSBezierPath*	path = [NSBezierPath bezierPath];
    int	count = [_anchorPoints count];
    int	i;
    
    if (count >= 1) {
        [path moveToPoint:[self pointAtIndex:0]];
        
        if (count >= 2) {
            for (i = 1; i < [_anchorPoints count]; i++) {
                [path lineToPoint:[self pointAtIndex:i]];
            }
        }
        
        if (_isPathClosed) {
            [path closePath];
        }
        
    }
    
    [path setLineWidth:[self strokeLineWidth]];
    
    return path;
}

- (void)drawInView:(SKTGraphicView *)view isSelected:(BOOL)flag {
    NSBezierPath *path = [self bezierPath];
    if (path) {
        if ([self drawsFill]) {
            [[self fillColor] set];
            [path fill];
        }
        if ([self drawsStroke]) {
            [[self strokeColor] set];
            [path stroke];
        }
    }
    if (flag) {
        [self drawHandlesInView:view];
    }
    
    // Handles for editing
    if ([self isCreating] || [self isEditing]) {
        // Draw handles for anchors
        [self drawAnchorHandlesInView:view];
    }
}

- (void)drawAnchorHandlesInView:(SKTGraphicView*)view {
    int	i;
    NSPoint	point;
    NSRect	rect;
    BOOL	isCreating = [self isCreating];
    BOOL	isEditing = [self isEditing];
    
    for (i = 0; i < [_anchorPoints count]; i++) {
        point = [self pointAtIndex:i];
        
        if (isCreating) {
            rect.origin.x = point.x - SKTPolygonHandleWidthHalf;
            rect.origin.y = point.y - SKTPolygonHandleWidthHalf;
            rect.size.width = SKTPolygonHandleWidthHalf * 2 + 1.0;
            rect.size.height = SKTPolygonHandleWidthHalf * 2 + 1.0;
            
            [[NSColor blackColor] set];
            NSRectFill(rect);
        }
        else if (isEditing) {
            rect.origin.x = point.x - SKTPolygonHandleWidthHalf;
            rect.origin.y = point.y - SKTPolygonHandleWidthHalf;
            rect.size.width = SKTPolygonHandleWidthHalf * 2 + 1.0;
            rect.size.height = SKTPolygonHandleWidthHalf * 2 + 1.0;
            
            [[NSColor whiteColor] set];
            NSRectFill(rect);
            
            [[NSColor blackColor] set];
            NSFrameRect(rect);
            
            if (i == _selectedPointIndex) {
                NSFrameRect(NSInsetRect(rect, 1.0, 1.0));
            }
        }
    }
}

//
// Event handling
//

- (BOOL)createWithEvent:(NSEvent *)event inView:(SKTGraphicView *)view {
	[self setLineArrowMode:[view lineArrowMode]];
	[self setFillColor:[view fillColor]];
	[self setDrawsFill:[view drawsFill]];
	[self setStrokeColor:[view strokeColor]];
	[self setDrawsStroke:[view drawsStroke]];
	[self setStrokeLineWidth:[view strokeLineWidth]];

    NSPoint	point = [view convertPoint:[event locationInWindow] fromView:nil];
    NSRect	bounds, drawingBounds;
    BOOL	snapsToGrid = [view snapsToGrid];
    float	spacing = [view gridSpacing];
    BOOL	echoToRulers = [[view enclosingScrollView] rulersVisible];
    NSEventType	type;
    
    [self startBoundsManipulation];
    if (snapsToGrid) {
        point.x = floor((point.x / spacing) + 0.5) * spacing;
        point.y = floor((point.y / spacing) + 0.5) * spacing;
    }
    [self setBounds:NSMakeRect(point.x, point.y, 0.0, 0.0)];
    if (echoToRulers) {
        [view beginEchoingMoveToRulers:[self bounds]];
    }
    
    [self appendPoint:point];
    
    // Tracking loop
    while (1) {
        event = [[view window] nextEventMatchingMask:(NSLeftMouseDownMask | NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        point = [view convertPoint:[event locationInWindow] fromView:nil];
        if (snapsToGrid) {
            point.x = floor((point.x / spacing) + 0.5) * spacing;
            point.y = floor((point.y / spacing) + 0.5) * spacing;
        }
        
        type = [event type];
        if (type == NSLeftMouseDown && [event clickCount] <= 1) {
            int	index = [self anchorPointUnderPoint:point];
            if (index == 0) {
                // Close path and go out from loop
                _isPathClosed = YES;
                break;
            }
            
            // Append point for a bezier path
            [self appendPoint:point];
        }
        else if (type == NSLeftMouseUp || type == NSLeftMouseDragged) {
            // Move the last point
            [self movePointAtIndex:[_anchorPoints count] - 1 toPoint:point];
        }
        
        // Set bounds and calculate drawing bounds
        drawingBounds = [self drawingBounds];
        [self setBounds:[[self bezierPath] bounds]];
        drawingBounds = NSUnionRect([self drawingBounds], drawingBounds);
        [view setNeedsDisplayInRect:drawingBounds];
        
        if (echoToRulers) {
            [view continueEchoingMoveToRulers:[self bounds]];
        }
        
        // In case of double click
        if ([event clickCount] >= 2) {
            break;
        }
    }
    if (echoToRulers) {
        [view stopEchoingMoveToRulers];
    }
    
    [self stopBoundsManipulation];
    
    bounds = [self bounds];
    if ((bounds.size.width > 0.0) || (bounds.size.height > 0.0)) {
        return YES;
    } else {
        return NO;
    }
}

- (BOOL)isEditable {
    return YES;
}

- (void)startEditingWithEvent:(NSEvent *)event inView:(SKTGraphicView *)view {
    [super startEditingWithEvent:event inView:view];
    
    [self applyTransform];
    [view setEditingGraphic:self editorView:nil];
    [view setNeedsDisplayInRect:[self drawingBounds]];
    
    return;
}

- (void)endEditingInView:(SKTGraphicView *)view {
    [super endEditingInView:view];
    
    _selectedPointIndex = -1;
    return;
}

- (BOOL)editWithEvent:(NSEvent*)event inView:(SKTGraphicView*)view {
    NSPoint	point = [view convertPoint:[event locationInWindow] fromView:nil];
    int		pointIndex = [self anchorPointUnderPoint:point];
    BOOL	snapsToGrid = [view snapsToGrid];
    float	spacing = [view gridSpacing];
    BOOL	echoToRulers = [[view enclosingScrollView] rulersVisible];
    
    if (pointIndex < 0) {
        return NO;
    }
    
    _selectedPointIndex = pointIndex;
    while (1) {
        event = [[view window] nextEventMatchingMask:(NSLeftMouseDraggedMask | NSLeftMouseUpMask)];
        
        if ([event type] == NSLeftMouseDragged) {
            point = [view convertPoint:[event locationInWindow] fromView:nil];
            if (snapsToGrid) {
                point.x = floor((point.x / spacing) + 0.5) * spacing;
                point.y = floor((point.y / spacing) + 0.5) * spacing;
            }
            
            [self movePointAtIndex:pointIndex toPoint:point];
        }
        
        [self setBounds:[[self origBezierPath] bounds]];
        [view setNeedsDisplayInRect:[self drawingBounds]];
        if (echoToRulers) {
            [view continueEchoingMoveToRulers:[self bounds]];
        }
        
        if ([event type] == NSLeftMouseUp) {
            break;
        }
    }
    
    return YES;
}

- (void)deleteForEditingInView:(SKTGraphicView*)view {
    NSRect	drawingBounds;
    
    if (_selectedPointIndex != -1) {
        // If the points is last one
        if ([_anchorPoints count] == 1) {
            [[self document] removeGraphic:self];
            return;
        }
        
        drawingBounds = [self drawingBounds];
        
        [self removePointAtIndex:_selectedPointIndex];
        _selectedPointIndex = -1;
        
        [self setBounds:[[self origBezierPath] bounds]];
        drawingBounds = NSUnionRect(drawingBounds, [self drawingBounds]);
        [view setNeedsDisplayInRect:drawingBounds];
    }
}

//
// Polygon specific
//

- (int)anchorPointUnderPoint:(NSPoint)point {
    int	i;
    NSPoint	pt;
    NSRect	rect;
    
    for (i = 0; i < [_anchorPoints count]; i++) {
        pt = [self pointAtIndex:i];
        pt = [_transform transformPoint:pt];
        
        rect.origin.x = pt.x - SKTPolygonHandleWidthHalf;
        rect.origin.y = pt.y - SKTPolygonHandleWidthHalf;
        rect.size.width = SKTPolygonHandleWidthHalf * 2 + 1.0;
        rect.size.height = SKTPolygonHandleWidthHalf * 2 + 1.0;
        
        if (NSPointInRect(point, rect)) {
            return i;
        }
    }
    
    return -1;
}

- (void)applyTransform {
    NSMutableArray*	anchorPoints = [NSMutableArray arrayWithCapacity:[_anchorPoints count]];
    NSEnumerator*	enumerator = [_anchorPoints objectEnumerator];
    id				object;
    NSPoint			point;
    
    while (object = [enumerator nextObject]) {
        point = [object pointValue];
        point = [_transform transformPoint:point];
        [anchorPoints addObject:[NSValue valueWithPoint:point]];
    }
    
    // Swap anchor points
    [_anchorPoints release];
    _anchorPoints = [anchorPoints retain];
    
    // Normarize transform
    [_transform setTransformStruct:SKTNormalizeStruct];
    
    _isFlippedVertically = NO;
    _isFlippedHorizontally = NO;
}

@end
